/* This file consists of functions which are called by the generated code
   to do I/O operations. */

/* Need to modify this to add/delete updated_tiles and cached_tiles
 * depending on whether or not dynamic cache is activated. */
#include <stdio.h>
#include "mapper.h"
#include "io.h"
#include "globals.h"
#include "consts.h"

#define dprintf_PPU if (1 && global_debug) printf

struct nes_ppu ppu;

int vbl=0;
int hvscroll=0;
int sprite0hit;

/* Declaration of global variables */
unsigned char vram[16384];
unsigned int controller1=0,controller2=0;
unsigned char spriteram[256];
unsigned int hvmirror=0;
unsigned int nomirror=0;
unsigned int osmirror=0;
unsigned int CLOCK; /* Current scanline position */
unsigned char hscrollreg,hscrollreg_hold, vscrollreg;

int updated_tiles[512];
int next_tile_to_update = 0;
int cached_tiles[512];

/* This is called whenever the game reads from 2xxx or 4xxx */
void input(int addr)
{
  static signed char vram_read;
  unsigned char *ptr;
  INRET=0;
  
  /* Read PPU status register */
  if(addr==0x2002)
  {
    /* FIXME Ewww... */
    sprite0hit=spriteram[0]*341+341+85+spriteram[3]+40; /* +40 is to account for the latency of the PPU; from the time it reads the background tile to the time the flag is set is about 40 cycles, give or take a few. */
    if(RAM[0x2000]&0x20) ptr=VRAM+((spriteram[1]&0xFE)<<4)+((spriteram[1]&1)<<12); /* 8x16 sprites */
    else ptr=VRAM+(spriteram[1]<<4)+((RAM[0x2000]&0x08)<<9); /* 8x8 sprites */
    if((RAM[0x2000]&0x20)&&(((long long *)ptr)[0]|((long long *)ptr)[8])==0)
      {sprite0hit+=8*341;ptr+=16;}
    while((ptr[0]|ptr[8])==0&&sprite0hit<spriteram[0]*341+5797) {sprite0hit+=341;ptr++;}
    if((vbl&&CLOCK>=VBL&&CLOCK<29695)|(CLOCK<VBL&&CLOCK>27393)) {INRET|=(signed char)0x80;vbl--;} /* I'm sure this isn't really accurate.  The vblank flag is set before the NMI is triggered, but the timing here is just a guess; also a read should always clear the flag. */
    if((CLOCK*3>=sprite0hit)&&(CLOCK<29695)) INRET|=(signed char)0x40;
    /*if((CLOCK*3>=sprite0hit)&&(CLOCK<VBL)) INRET|=(signed char)0x40;  /* When does sprite0hit go off? */
  }
  
  /* Read from VRAM */
  if(addr==0x2007)
  {
    INRET=vram_read;
	/* FIXME Ewww... use pointers */
    if(osmirror&&VRAMPTR>=0x2000&&VRAMPTR<=0x2FFF) vram_read=VRAM[VRAMPTR&0x23FF]; /* one screen */
    else if(VRAMPTR>0x3f00) vram_read=VRAM[VRAMPTR&0x3f1f];
    else if(!nomirror&&hvmirror&&VRAMPTR>=0x2400&&VRAMPTR<0x2c00) vram_read=VRAM[VRAMPTR-0x400];
    else vram_read=VRAM[VRAMPTR];
    VRAMPTR+=1<<(((*((unsigned char *)REG1)&4)>>2)*5); /* bit 2 of $2000 controls increment */
    if(VRAMPTR>0x3FFF) VRAMPTR&=0x3fff;
    /*printf("VRAM Read: %4x=%2x (scan %d)\n",VRAMPTR,INRET,CLOCK);/* */

    /* This is the so-called mid-hblank update.  If an address is written
       into 2006 followed by reads of 2007 during refresh, then the
       address is loaded into the PPU and used as the start address of the
       next scanline.  This is used to scroll the background vertically
       on a portion of the screen.  We need to convert the scanline
       address into horizontal/vertical offsets. */
    if(CLOCK<VBL&&(RAM[0x2001]&8)!=0)
    {
   		int val; 
	  dprintf_PPU("* ppu read from 2007\n");
      drawimage(CLOCK*3);
      vline=(VRAMPTR&0x3e0)>>5;
	  /* vscan = 0?  why not VRAMPTR >> 12? */
      vscan=0;
	  /*vscan = VRAMPTR >> 12; */
    if(osmirror) ppu.scanpage=0x2000;
    else if(nomirror) ppu.scanpage=0x2000 + (VRAMPTR & 0x0c00);
    else if(hvmirror==0) ppu.scanpage=0x2000+(VRAMPTR & 0x0400);
    else if(hvmirror==1) ppu.scanpage=0x2000+((VRAMPTR & 0x0800) >> 1);
	ppu.scanpage_temp = ppu.scanpage;

	dprintf_PPU("   new 2000 scanpage: %04x  osmirr: %d hvmirr: %d vwrap: %d\n", ppu.scanpage, osmirror, hvmirror, vwrap);
		dprintf_PPU("      new scanpage: %04x\n", ppu.scanpage);   
      if(hvmirror||nomirror) vwrap=((VRAMPTR&0x800)>>11)^((RAM[0x2000]&2)>>1);

      /*printf("Read: %4x (scan %d, sprite0 %d) 2000=%2x 2001=%2x vbl=%d\n",addr,CLOCK,sprite0hit*HCYCLES,RAM[0x2000],RAM[0x2001],vbl);/* */

#define DEBUG_VSCROLL
#ifdef DEBUG_VSCROLL
	  dprintf_PPU("      new vline: %d vscan: %d vwrap: %d scanline: %d\n", vline, vscan, vwrap, CLOCK*3/341);
#endif
#undef DEBUG_VSCROLL

    }
  }

  /* Read joypad controller */
  if(addr==0x4016)
  {
    INRET=(RAM[0x4016]&1);
    RAM[0x4016]>>=1;
  }

  /* debug stuff:
  printf("Read: %4x=%2x (scan %d)\n",addr,INRET&0xff,CLOCK);/*
  printf("\nstack: %x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x\n",
    (&addr)[1],(&addr)[2],(&addr)[3],(&addr)[4],(&addr)[5],(&addr)[6],
    (&addr)[7],(&addr)[8],(&addr)[9],(&addr)[10],(&addr)[11],(&addr)[12] );
  /**/
}

/* This is called whenever the game writes to 2xxx or 4xxx */
void output(int addr,int val)
{
  static int spriteaddr;  /* FIXME */

  val&=0xFF;

  /* Select pattern table */
  if(addr==0x2000)
  {
	dprintf_PPU("* ppu write to 2000 val %02x\n", val);
    drawimage(CLOCK*3);
	/* LOOKHERE */

	/* scanpage_temp was introduced because changing pattern table is almost equivalent to changing
	/* scroll registers--and vscroll registers can't be changed outside of VBL.  But hscroll regs
	/* can--note hvmirror == 0 case.  */
	ppu.scanpage_temp = 0x2000 + ((val & 3) << 10);
    if(osmirror) ppu.scanpage_temp=0x2000;
    else if(nomirror) ppu.scanpage_temp=0x2000 + ((val & 3) << 10);
		/* Name table address in lower 2 bits of $2000 is used as-is. */
    else if(hvmirror==0) { ppu.scanpage_temp=0x2000+((val&1)<<10); ppu.scanpage = ppu.scanpage_temp; }
		/* lower bit accepted; allowing 0x2000 and 0x2400; 0x2800 mirrors
		 * 0x2000, 0x2c00 mirrors 0x2400.  Change is available next scanline */
    else if(hvmirror==1) ppu.scanpage_temp=0x2000+((val&2)<<9);
		/* upper bit accepted but used as lower addr bit; 0x2000 and
		 * 0x2800 allowed input; 0x2400 and 0x2000 are mapped to 0x2000;
		 * 0x2800 and 0x2c00 are mapped to 0x2000.  This is because emu
		 * stores data internally at 0x2000 and 0x2400 (stupid non-use of
		 * pointers).  This is horizontal mirroring, where 0x2800 mirrors
		 * 0x2c00, but the page for 0x2800 and 0x2c00 is stored at 0x2400.
		 * Needs to be fixed to use pointers. */
		/* However, I don't understand why Gauntlet _has_ four screens avail.--
		 * perhaps only the nomirror case stores data interally in all scanpages? */

	dprintf_PPU("   new scanpage (pending): %04x  osmirr: %d hvmirr: %d vwrap: %d\n", ppu.scanpage_temp, osmirror, hvmirror, vwrap);

    RAM[0x2000]=val;
  }

  if(addr==0x2001)
  {
	dprintf_PPU("* ppu write to 2001 val %02x\n", val);
    drawimage(CLOCK*3);
    RAM[0x2001]=val;
  }

  /* Set horizontal/vertical scroll */
  /* Require two bytes be sent (first is horiz; second is vert.) before update takes effect */
  if(addr==0x2005)
  {
	dprintf_PPU("* ppu write to 2005 val %02x clock %d (l%d/h%d)\n", val, CLOCK*3, CLOCK*3/341,CLOCK*3%341);
    if(hvscroll^=1)
    {
      hscrollreg_hold=val;
    }
    else
    {
	  dprintf_PPU("   pre-draw hscroll %d\n", hscrollreg);
      drawimage(CLOCK*3);
	  hscrollreg = hscrollreg_hold;
      vscrollreg=val;
	  dprintf_PPU("   post-draw hscroll %d\n", hscrollreg);
    }
  }
  
  /* I think we should not update this register until two values are
   * written */
  /* Also scanpage should not? be updated until subsequent read from 2007? */
   
  /* Load VRAM target address */
  if(addr==0x2006)
  {
	dprintf_PPU("* ppu write to 2006 val %02x clock %d (l%d/h%d)\n", val, CLOCK*3, CLOCK*3/341,CLOCK*3%341);
    drawimage(CLOCK*3);
    VRAMPTR=((VRAMPTR&0x3f)<<8)|(val&0xff);

	/* I believe that, during vblank, writes to 2006 can update the PPU
	 * address, but the only thing you can do is set the scanpage, since
	 * a scanline change must be effected during mid-hblank (equivalent
	 * effect during vblank is setting the vscrollreg--via 2005.) */
	/* This is probably wrong; need to copy scanpage settings from 0x2000. */
	if (CLOCK>=VBL) {
		ppu.scanpage_temp = 0x2000 + (VRAMPTR & 0x0c00);
		dprintf_PPU("      new scanpage: %04x\n", ppu.scanpage_temp); 
	}
    /* For mid-hblank updates: */
    /* Note: When the VRAM address register is loaded, the scanline number
       is derived from the top bits of the address.  When the address
       register load is followed by a read of 2005, then the scanline
       number is reset to zero and the scan starts with the top of the
       current char/tile line. (see above) */
/*    vline=(VRAMPTR&0x3e0)>>5;
    vscan=VRAMPTR>>12; */
    if(hvmirror||nomirror) vwrap=((VRAMPTR&0x800)>>11)^((RAM[0x2000]&2)>>1);
/*	dprintf_PPU("      new vline: %d vscan: %d vwrap: %d\n", vline, vscan, *	vwrap); */
    /*if(CLOCK<VBL)printf("hbl update: %4x %d\n",VRAMPTR,CLOCK);*/
    hvscroll=0;
  }

  /* Write VRAM */
  if(addr==0x2007)
  {
    /* A little bit about VRAM:
       
       The NES has two pages of internal VRAM.  The emulator always stores
       the data internally at 2000 and 2400, however the page at 2400 may
       be mapped to 2800 in the PPU address space, with the page at 2400
       mirroring 2000, or vice versa.
       
       0000-1FFF is mapped in from the game cartridge, and may be RAM or ROM.
       
       3Fxx holds the color pallette.
    */

    /*mmc2_latch(VRAMPTR);*/
    /*if(CLOCK<VBL&&(RAM[0x2001]&8)) printf("vram write during refresh! "); /* For debugging */
    /*printf("VRAM: %4x=%2x (+%d) scan %d\n",VRAMPTR,val,1<<(((*((unsigned char *)REG1)&4)>>2)*5),CLOCK); /* */
	/* Palette updating code elsewhere takes background color from 0x3f00.  Code here
	 * therefore doesn't bother mirroring 3f00 to 04, 08, 0c, etc.  but does mirror
	 * it to 3f10 (sprite bg).  OK, except writes to 04 or 08 etc. should affect 3f00,
	 * right?  Not done here. */
    if((VRAMPTR&0x3f0f)==0x3f00)
      VRAM[0x3f00]=VRAM[0x3f10]=(unsigned char)val; /* Background color is mirrored between pallettes */
    else if((VRAMPTR&0x3f00)==0x3f00) {
      VRAM[VRAMPTR&0x3f1f]=(unsigned char)val; /* Write to color pallette */
/*	  if ((VRAMPTR&0x3f1f) < 0x3f10) printf("tile palette change: %04x -> %02x\n", VRAMPTR&0x3f1f, val);
	  else if ((VRAMPTR&0x3f1f) > 0x3f10) printf("sprite palette change: %04x -> %02x\n", VRAMPTR&0x3f1f, val); */
	  memset(cached_tiles, 0, sizeof(int) * 512); 
	}
    else if(VRAMPTR>=0x2000&&VRAMPTR<0x3000)
    {
	  /* FIXME This is where incoming value is translated to emu's two internal pages at 2000 and 2400,
	   * unless there's no mirroring in which case we write directly to VRAM. */
      if(nomirror)
        VRAM[VRAMPTR]=(unsigned char)val;
      else if(osmirror)
        VRAM[VRAMPTR&0x23FF]=(unsigned char)val; /* One-Screen Mirroring */
      else if(!hvmirror)
        VRAM[VRAMPTR]=VRAM[VRAMPTR^0x800]=(unsigned char)val;
      else if(hvmirror)
      {
        if(VRAMPTR>=0x2000&&VRAMPTR<0x2400) VRAM[VRAMPTR]=VRAM[VRAMPTR^0x800]=(unsigned char)val;
        if(VRAMPTR>=0x2400&&VRAMPTR<0x2800) VRAM[VRAMPTR-0x400]=VRAM[VRAMPTR+0x400]=(unsigned char)val;
        if(VRAMPTR>=0x2800&&VRAMPTR<0x2c00) VRAM[VRAMPTR-0x400]=VRAM[VRAMPTR+0x400]=(unsigned char)val;
        if(VRAMPTR>=0x2c00&&VRAMPTR<0x3000) VRAM[VRAMPTR]=VRAM[VRAMPTR^0x800]=(unsigned char)val;
      }
    }
    else {
		/* Access between 0x0000 and 0x1fff, to pattern table. */
		VRAM[VRAMPTR]=(unsigned char)val;
		/* 16 bytes == 8 bytes per color plane * 2 color planes per tile */
		cached_tiles[VRAMPTR / 16] = 0;
	}
    
	/* FIXME what is this REG1 stuff */
    VRAMPTR+=1<<(((*((unsigned char *)REG1)&4)>>2)*5); /* bit 2 of $2000 controls increment */
	/* FIXME this really should be inbounds */
    if(VRAMPTR>0x3FFF){ /*printf("help - vramptr >0x3fff!\n");/*exit(1);*/VRAMPTR&=0x3fff;}
  }
  
  /* Sprite DMA */
  if(addr==0x4014)
    /* I'm not entirely sure of this, but it seems accurate. */
    {
      drawimage(CLOCK*3);
      if(val<0x80)
        memcpy(spriteram,RAM+(val<<8),256);
      else
        memcpy(spriteram,MAPTABLE[val>>4]+(val<<8),256);
      CLOCK+=514;
      CTNI+=514;
    }

  if(addr==0x2003) spriteaddr=val;
  if(addr==0x2004) spriteram[spriteaddr]=val;

  /* Reset controller */
  if((addr|1)==0x4017) {
    RAM[0x4016]=controller1;
    RAM[0x4017]=controller2;
  }

  /* more debugging stuff:
  printf("Write: %4x,%2x (scan %d)\n",addr,val,CLOCK);/*
  printf("stack: %x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x\n",
    (&addr)[1],(&addr)[2],(&addr)[3],(&addr)[4],(&addr)[5],(&addr)[6],
    (&addr)[7],(&addr)[8],(&addr)[9],(&addr)[10],(&addr)[11],(&addr)[12] );
  /**/

}

/* This determines whether an NMI should occur and returns 1 if so. */
/* This function is called even when interrupts are off, and also */
/* refreshes the screen as necessary. */
int donmi()
{

  /*printf("donmi: at %d\n",CLOCK);*/

  CLOCK=VBL+7; /* 7 cycle interrupt latency */
  CTNI=-CPF+7;
  vbl=1;

  UpdateDisplay();

  /*printf("donmi: stack at %x\n",STACKPTR);*/

  /* reset scroll registers */
  hvscroll=0;

  /* Is an NMI to be generated?  Return 1 if so. */
  if(((*((unsigned char *)(REG1)))&0x80)!=0) return 1;
  else return 0;
}

/* this is left in here for debugging purposes but is not normally called: */
trace(int s)
{
/*
  printf("branch, stack: %x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x\n",
    (&s)[1],(&s)[2],(&s)[3],(&s)[4],(&s)[5],(&s)[6],
    (&s)[7],(&s)[8],(&s)[9],(&s)[10],(&s)[11],(&s)[12] );
*/
  unsigned int x;
  char hex[17]="0123456789ABCDEF";
  x=((int *)(&s))[4];
  printf("%c%c%c%c\n",hex[x>>12],hex[(x&0xf00)>>8],hex[(x&0xf0)>>4],hex[x&0xf]);
  fflush(stdout);
}
