#undef X11_DRAW
#undef SLOW_DOWN_IF_AHEAD
#undef SPEED_UP_IF_BEHIND

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xutil.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "mapper.h"
#include "io.h"
#include "globals.h"
#include "consts.h"


/* X11 stuff: */
Display *display;
Visual *visual;
Window rootwindow,w;
int screen;
Colormap colormap;
XEvent ev;
unsigned char *keystate[32];
GC gc,blackgc;
GC addgc,maskgc;
GC backgroundgc,solidbggc;
GC color1gc,color2gc,color3gc,bgcolorgc;
GC spritesolidcolorgc,spritecolor1gc,spritecolor2gc,spritecolor3gc;
GC planegc,blackplanegc,planenorgc,planeandgc,planexorgc;
int black,white;
XSizeHints sizehints;
XGCValues GCValues;
XSetWindowAttributes windowattrib;
unsigned long colortable[25];
int indexedcolor=0;
XColor color;

#define tilecachedepth 25

Pixmap background[tilecachedepth];  /* The pre-rendered background that sprites are drawn on top of */
Pixmap bgplane1[tilecachedepth],bgplane2[tilecachedepth],bgplane3[tilecachedepth],bgmask[tilecachedepth];  /* Bitmaps for each color plane of the rendered background */
Pixmap tilecache1[tilecachedepth],tilecache2[tilecachedepth],tilecache3[tilecachedepth],tilebgmask[tilecachedepth];  /* Bitmaps for the background tiles in each of the four colors */
Pixmap layout; /* To assemble the final image to be displayed */
XImage *tilebitimage,*tilebitimage2;
unsigned char tilebitdata[16384],tilebitdata2[16384];
XImage *image;

/* Flags for differential redraw */
unsigned char needsredraw=1; /* Refresh screen display */
unsigned char redrawbackground=1; /* Redraw tile background */
unsigned char redrawall=1; /* Redraw all scanlines */
unsigned char bgmask_changed[tilecachedepth];
unsigned char current_bgmask=0;
unsigned char tilebgmask_changed[tilecachedepth];
unsigned char tilemask1_changed[tilecachedepth];
unsigned char tilemask2_changed[tilecachedepth];
unsigned char tilemask3_changed[tilecachedepth];
unsigned char scanline_diff[240];
unsigned short int tileline_begin[60];
unsigned short int tileline_end[60];

/* Precalculated/Cached values */

/* Misc globals */
int hoffset,voffset;
int oldhoffset,oldvoffset;
unsigned int currentline;
int debug_bgoff=0;
int debug_spritesoff=0;
int debug_switchtable=0;
int basetime;
int desync=1;
int magstep=1;
int currentcache=0;
int nextcache=1%tilecachedepth;

GrTexInfo bgtile_info;
FxU32 bgtile_start;
FxU32 bgtile_size;
int pause_emu=0;
int sleep_emu=0;

void update_colors();

#define screen_on (RAM[0x2001]&8)
#define sprites_on (RAM[0x2001]&16)
#define maxsize 1 /* Can't resize window yet */


/* 6-bit nes color palette--total 64 colors, but 12 are duplicates,
 * so 52 unique colors.  Derivation:
   bit format 000llhhhh     (l = lum, h = hue)
   2^2 = 4 brightness values * 2^4 = 16 color values == 64 colors.
   However only hues 0-12 are valid (13,14,15 are duplicates of 0 [black])
   so 4 brightness * 13 color = 52 colors.
   Note that hues 13,14,15 below are always represented as 0 [black] regardless
   of luminance value.  I don't know if this is correct.
 */

/* Here is the color palette.  Change it if you don't like it. */
/* I don't understand why hue 0/lum 0 is represented as 0x3f3f3f which is grey25. */
/* Does palette entry 0 do something I don't know about? */
unsigned int palette_24[64]={
  0x3f3f3f,0x001f3f,0x00003f,0x1f003f,0x3f003f,0x3f0020,0x3f0000,
    0x3f2000,0x3f3f00,0x203f00,0x003f00,0x003f20,0x003f3f,0,0,0,
  0x7f7f7f,0x405f7f,0x40407f,0x5f407f,0x7f407f,0x7f4060,0x7f4040,
    0x7f6040,0x7f7f40,0x607f40,0x407f40,0x407f60,0x407f7f,0,0,0,
  0xbfbfbf,0x809fbf,0x8080bf,0x9f80bf,0xbf80bf,0xbf80a0,0xbf8080,
    0xbfa080,0xbfbf80,0xa0bf80,0x80bf80,0x80bfa0,0x80bfbf,0,0,0,
  0xffffff,0xc0dfff,0xc0c0ff,0xdfc0ff,0xffc0ff,0xffc0e0,0xffc0c0,
    0xffe0c0,0xffffc0,0xe0ffc0,0xc0ffc0,0xc0ffe0,0xc0ffff,0,0,0,
  };
unsigned short palette_16[64];

#include <glide.h>
#include <assert.h>

unsigned int ttlframes;
struct timeval btv;

void display_palette();

void display_palette() {
	int i;

	for (i = 0x3f00; i <= 0x3f1f; i++) {
		if (i%8 == 0) {
			printf("\n%04x: ", i);
		}
		printf("%02x(%06x) ", VRAM[i], palette_16[VRAM[i]]);
	}
	printf("\n");
	printf("palette[] ");

	for (i = 0; i < 24; i++) {
		printf("%d(%06x) ", i, ((unsigned short int *)palette)[i]); 
	}
	printf("\n");
}

int init_glide_display() {
    GrScreenResolution_t resolution = GR_RESOLUTION_640x480;
	char version[80];
    GrHwConfiguration hwconfig;
    grGlideGetVersion( version );
    printf( "Initializing Glide version %s\n", version );
    grGlideInit();
    assert(
		grSstQueryHardware( &hwconfig )
	);
    grSstSelect( 0 );
    assert(
		grSstWinOpen( 0, resolution, GR_REFRESH_60Hz,
	        GR_COLORFORMAT_ABGR, GR_ORIGIN_UPPER_LEFT, 2, 1 )
	);

    grColorCombine( GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_ONE,
                    GR_COMBINE_LOCAL_NONE, GR_COMBINE_OTHER_TEXTURE, FXFALSE );
    grTexCombine( GR_TMU0, GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE,
                  GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE,
                  FXFALSE, FXFALSE );
	grAlphaCombine( GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_ONE,
	                GR_COMBINE_LOCAL_NONE, GR_COMBINE_OTHER_TEXTURE, FXFALSE ); 
    grAlphaTestReferenceValue(0x30);
    grAlphaTestFunction(GR_CMP_GREATER);  

    grTexMipMapMode( GR_TMU0, GR_MIPMAP_DISABLE, FXFALSE );
	grClipWindow(0, 0, BG_TILE_SCRW * 32, BG_TILE_SCRH * 30);

	grDepthBufferMode(GR_DEPTHBUFFER_ZBUFFER);
	grDepthBufferFunction(GR_CMP_GREATER);
	grDepthMask(FXTRUE);

	/*grTexFilterMode( GR_TMU0, GR_TEXTUREFILTER_BILINEAR,
	 * GR_TEXTUREFILTER_BILINEAR ); */

	printf("F1-F4 - Mini displays\n");
	printf("F5 - Bilinear filtering   F6 - Point Sampling\n");
	printf("F7 - Pause emulator       F8 - Frame advance\n");
	printf("F9 - Turn off background  F10 - Turn off sprites\n");
	printf("d  - debugging messages   p - palette\n");
	printf("t - tile display  y - tile display color (0-3)\n");
    printf( "Initialized.\n" );

/*bzero(updated_tiles, 512 * sizeof(int));   */
/* updated_tiles: obsolete cache list containing the numbers of all
 * modified tiles since last vblank.  Renderer updates all colors of all
 * tiles in the queue, regardless of whether or not those tiles are used.
 * This means 2048 tiles are updated per palette switch.
*/
/*	memset(updated_tiles, 1, 512 * sizeof(int));   */

/* cached_tiles: an array with one element for each tile.  It's a bitfield
 * with bit n set if that "color" of the tile (i.e. attribute permutation)
 * is in memory, clear if not.  When the renderer tries to draw a tile,
 * array is checked; the desired tile color is put into texture memory
 * if necessary and then drawn.  A tile's entry is cleared when its
 * corresponding VRAM is modified so it will be reloaded on next use.
 * The entire array is invalidated when the palette is updated, so
 * that all tiles will be reloaded.  */
	memset(cached_tiles, 0, 512 * sizeof(int));

/*  Initialize the generic background tile GrTexInfo structure, and get memory
    start and size values for the textures. */
	bgtile_info.smallLod = GR_LOD_8;
	bgtile_info.largeLod = GR_LOD_8;
	bgtile_info.aspectRatio = GR_ASPECT_1x1;
	bgtile_info.format	= GR_TEXFMT_ARGB_1555;
    bgtile_start = grTexMinAddress( GR_TMU0 );
    bgtile_size  = grTexTextureMemRequired(GR_MIPMAPLEVELMASK_BOTH, &bgtile_info);

	global_display_glide_bg = 1;
	global_display_glide_sprites = 1;
	global_display_mini1 = 0;
	global_display_mini2 = 0;
	global_display_mini3 = 0;
	global_display_mini4 = 0;
	global_display_all_tiles = 0;
	global_display_all_tiles_col = 0;
	global_debug = 0;

	return 1;
}

InitDisplay()
{
  int x;
  struct timeval time;

  if (!init_glide_display()) return;

  display=XOpenDisplay(NULL);
  if(display==NULL) return;
  screen=XDefaultScreen(display);
  visual=XDefaultVisual(display,screen);
  rootwindow=XRootWindow(display,screen);
  colormap=DefaultColormap(display,screen);
  white=WhitePixel(display,screen);
  black=BlackPixel(display,screen);
  depth=DefaultDepth(display,screen);
  if(depth==8)
  {
    indexedcolor=1;
    if(XAllocColorCells(display,colormap,0,0,0,colortable,25)==0)
    {
      printf("Can't allocate colors!\n");
      exit(1);
    }
    /* Pre-initialize the colormap to known values */
    color.red=((palette_24[0]&0xFF0000)>>8);
    color.green=(palette_24[0]&0xFF00);
    color.blue=((palette_24[0]&0xFF)<<8);
    color.flags=DoRed|DoGreen|DoBlue;
    for(x=0;x<=24;x++)
    {
      ((unsigned char *)palette)[x]=color.pixel=colortable[x];
      XStoreColor(display,colormap,&color);
    }
  }
  if(depth==16)
  {
    /* Create 16-bit pallette */
    for(x=0;x<64;x++)
    {
      palette_16[x]=((palette_24[x]>>3)&0x1f)|((palette_24[x]>>5)&0x7c0)
                     |((palette_24[x]>>8)&0xf800);
    }
  }
  w=XCreateSimpleWindow(display,rootwindow,0,0,256,240,0,black,black);
  gc=XCreateGC(display,w,0,0);
  blackgc=XCreateGC(display,w,0,0);
  GCValues.function=GXor;
  addgc=XCreateGC(display,w,GCFunction,&GCValues);
  GCValues.function=GXand;
  maskgc=XCreateGC(display,w,GCFunction,&GCValues);
  XSetForeground(display,gc,~0);
  XSetBackground(display,gc,0);
  XSetForeground(display,blackgc,0);
  XSetBackground(display,blackgc,0);
  GCValues.function=GXor;
  color1gc=XCreateGC(display,w,GCFunction,&GCValues);
  color2gc=XCreateGC(display,w,GCFunction,&GCValues);
  color3gc=XCreateGC(display,w,GCFunction,&GCValues);
  bgcolorgc=XCreateGC(display,w,GCFunction,&GCValues);
  XSetBackground(display,bgcolorgc,0);
  GCValues.function=GXcopy;
  spritesolidcolorgc=XCreateGC(display,w,GCFunction,&GCValues);
  spritecolor1gc=XCreateGC(display,w,GCFunction,&GCValues);
  spritecolor2gc=XCreateGC(display,w,GCFunction,&GCValues);
  spritecolor3gc=XCreateGC(display,w,GCFunction,&GCValues);

  /* set window title */
  XStoreName(display,w,"Nestra");
  /* set aspect ratio */
  sizehints.flags=PMinSize|PMaxSize|PResizeInc|PAspect|PBaseSize;
  sizehints.min_width=256;
  sizehints.min_height=240;
  sizehints.max_width=256*maxsize;
  sizehints.max_height=240*maxsize;
  sizehints.width_inc=256;
  sizehints.height_inc=240;
  sizehints.min_aspect.x=256;
  sizehints.min_aspect.y=240;
  sizehints.max_aspect.x=256;
  sizehints.max_aspect.y=240;
  sizehints.base_width=0;
  sizehints.base_height=0;
  XSetWMNormalHints(display,w,&sizehints);

  for(x=0;x<tilecachedepth;x++){
    background[x]=XCreatePixmap(display,w,512,480,depth); /* should make bigger for expanding window (fix this) */
    bgplane1[x]=XCreatePixmap(display,w,512,480,1);
    bgplane2[x]=XCreatePixmap(display,w,512,480,1);
    bgplane3[x]=XCreatePixmap(display,w,512,480,1);
    bgmask[x]=XCreatePixmap(display,w,512,480,1);
    tilecache1[x]=XCreatePixmap(display,w,256,64,1);
    tilecache2[x]=XCreatePixmap(display,w,256,64,1);
    tilecache3[x]=XCreatePixmap(display,w,256,64,1);
    tilebgmask[x]=XCreatePixmap(display,w,256,64,1);
  }
  planegc=XCreateGC(display,tilebgmask[0],0,0);
  XSetForeground(display,planegc,1);
  XSetBackground(display,planegc,0);
  blackplanegc=XCreateGC(display,tilebgmask[0],0,0);
  XSetForeground(display,blackplanegc,0);
  XSetBackground(display,blackplanegc,0);
  GCValues.function=GXnor;
  planenorgc=XCreateGC(display,tilebgmask[0],GCFunction,&GCValues);
  GCValues.function=GXand;
  planeandgc=XCreateGC(display,tilebgmask[0],GCFunction,&GCValues);
  GCValues.function=GXxor;
  planexorgc=XCreateGC(display,tilebgmask[0],GCFunction,&GCValues);
  backgroundgc=XCreateGC(display,w,0,0);
  XSetBackground(display,backgroundgc,black);
  solidbggc=XCreateGC(display,w,0,0);
  XSetBackground(display,solidbggc,black);
  layout=XCreatePixmap(display,w,256*maxsize,240*maxsize,depth);
  tilebitimage=XCreateImage(display,visual,1,XYBitmap,0,tilebitdata,256,64,8,32);
  tilebitimage2=XCreateImage(display,visual,1,XYBitmap,0,tilebitdata2,256,64,8,32);
  XMapWindow(display,w);
  XSelectInput(display,w,KeyPressMask|KeyReleaseMask|ExposureMask|FocusChangeMask|KeymapStateMask);
  XFlush(display);
  gettimeofday(&time,NULL);
  basetime=time.tv_sec;
  ttlframes=0;
  fbinit();
  if(depth==24)
    image=XCreateImage(display,visual,depth,ZPixmap,0,fb,256,240,32,0);
  else
    image=XCreateImage(display,visual,depth,ZPixmap,0,fb,256,240,depth,0);
}


UpdateDisplay()
{
  struct timeval time;
  static int first_time = 1;
  unsigned static int frame;
  unsigned int timeframe;
  static int halfspeed=0,doublespeed=0;
  static int advance_one_frame=0;

  if (first_time) { gettimeofday(&btv,NULL); first_time = 0; }
  
  /* Check the time.  If we're getting behind, skip a frame to stay in sync. */
  gettimeofday(&time,NULL);
  timeframe=(time.tv_sec-basetime)*50+time.tv_usec/20000; /* PAL */
  timeframe=(time.tv_sec-basetime)*60+time.tv_usec/16666; /* NTSC */
  frame++;
  if(halfspeed) timeframe>>=1;
  if(doublespeed==2) timeframe<<=1;
  else if(doublespeed>2) timeframe*=doublespeed;
  if(desync) frame=timeframe;desync=0;
  if(frame<timeframe-20&&frame%20==0) desync=1; /* If we're more than 20 frames behind, might as well stop counting. */
  if (advance_one_frame) {
  		sleep_emu = 1;
		advance_one_frame = 0;
		printf("\n");
  }

  /* Stupid addition to signal full frame update to tile renderer */
  endframe=1;
/*  printf("Update_display called\n"); */
  drawimage(81840);
/*  grBufferSwap(1); */
  endframe=1;
  if(!frameskip)
  {
    ttlframes++;
/*  printf("update_colors called\n");
    update_colors(); */
#ifdef X11_DRAW
    XPutImage(display,layout,gc,image,0,0,0,0,256,240);
#endif
/* bufferswap a little slower here than drawimage; a lot slower with idle. */
/*  grBufferSwap(1);  */
/*  grSstIdle();  */
    redrawall=needsredraw=1;
  }

  /* Slow down if we're getting ahead */
#ifdef SLOW_DOWN_IF_AHEAD
  if(frame>timeframe+1&&frameskip==0) {
    usleep(16666*(frame-timeframe-1));
  }
#endif
  
  /* Handle X input */
  while(XPending(display)||ev.type==-1||sleep_emu||pause_emu) {
    XNextEvent(display,&ev);
/*   printf("event %d\n",ev.type); */

	switch (ev.type) {
		case DestroyNotify:
			quit(); break;
		case FocusIn:
			break;
		case FocusOut:
			sleep_emu = desync = 1; break;
		case KeyPress:
			switch (((XKeyEvent *)&ev)->keycode) {
				case   9: quit(); break; /* ESC */ 
				case  36: controller1 |= STARTBUTTON; break; /* Return */ 
				case  23: controller1 |= SELECTBUTTON; break; /* Tab */ 
				case  98: controller1 |= UP; break;
				case 104: controller1 |= DOWN; break;
				case 100: controller1 |= LEFT; break;
				case 102: controller1 |= RIGHT; break;
				case  33: display_palette(); break; /* Z */ 
				case  52: controller1 |= BUTTONB; break; /* Z */ 
				case  53: controller1 |= BUTTONB; break; /* X */ 
				case  50: controller1 |= BUTTONB; break; /* Shift */ 
				case  65: controller1 |= BUTTONA; break; /* Space Bar */ 
				case  38: controller1 |= BUTTONA; break; /* A */ 
				case  40: global_debug ^= 1; break; /* D */ 
				case  54: controller1 |= BUTTONA; break; /* C */ 
				case  67: global_display_mini1 ^= 1; break; /* F1 */ 
				case  68: global_display_mini2 ^= 1; break; /* F2 */ 
				case  69: global_display_mini3 ^= 1; break; /* F3 */ 
				case  70: global_display_mini4 ^= 1; break; /* F4 */ 
				case  71: /* F5 */
						grTexFilterMode( GR_TMU0, GR_TEXTUREFILTER_BILINEAR,
						                 GR_TEXTUREFILTER_BILINEAR );
						break;
				case  72: /* F6 */
						grTexFilterMode( GR_TMU0, GR_TEXTUREFILTER_POINT_SAMPLED,
						                 GR_TEXTUREFILTER_POINT_SAMPLED);
						break;
				case  73: printf("\n"); sleep_emu ^= 1; break; /*F7*/ 
				case  74: sleep_emu = 0; advance_one_frame = 1; break; /*F8*/ 
				case  75: global_display_glide_bg ^= 1; break; /*F9*/  
				case  76: global_display_glide_sprites ^= 1;  break; /*F10*/  
				case 110: START(); break; /* Pause/Break */
				case  49: desync=1;halfspeed=1;doublespeed=0;pause_emu=0; break; /* ` */ 
				case  10: desync=1;halfspeed=0;doublespeed=0;pause_emu=0; break; /* 1 */ 
				case  11: desync=1;halfspeed=0;doublespeed=2;pause_emu=0; break; /* 2 */ 
				case  12: desync=1;halfspeed=0;doublespeed=3;pause_emu=0; break; /* 3 */ 
				case  13: desync=1;halfspeed=0;doublespeed=4;pause_emu=0; break; /* 4 */ 
				case  14: desync=1;halfspeed=0;doublespeed=5;pause_emu=0; break; /* 5 */ 
				case  15: desync=1;halfspeed=0;doublespeed=6;pause_emu=0; break; /* 6 */ 
				case  16: desync=1;halfspeed=0;doublespeed=7;pause_emu=0; break; /* 7 */ 
				case  17: desync=1;halfspeed=0;doublespeed=8;pause_emu=0; break; /* 8 */ 
				case  19: desync=1;pause_emu=1; break; /* 0 */ 
				case  28: global_display_all_tiles ^= 1; break; /* t */
				case  29: global_display_all_tiles_col++; global_display_all_tiles_col &= 3; break; /* t */
				
				default:
					break;
			}
			break;

		case KeyRelease:
			switch (((XKeyEvent *)&ev)->keycode) {
				case  36: controller1 &= ~STARTBUTTON; break; /* Return */
				case  23: controller1 &= ~SELECTBUTTON; break; /* Tab */
				case  98: controller1 &= ~UP; break;
				case 104: controller1 &= ~DOWN; break;
				case 100: controller1 &= ~LEFT; break;
				case 102: controller1 &= ~RIGHT; break;
				case  52: controller1 &= ~BUTTONB; break;
				case  53: controller1 &= ~BUTTONB; break;
				case  50: controller1 &= ~BUTTONB; break;
				case  65: controller1 &= ~BUTTONA; break;
				case  38: controller1 &= ~BUTTONA; break;
				case  54: controller1 &= ~BUTTONA; break;
			}
			break;

		case KeymapNotify:
			memcpy(keystate,((XKeymapEvent *)&ev)->key_vector,32);
			controller1 = controller2 = 0;
			break;
		case Expose:
			needsredraw = 1;
			break;
		default:
			break;
	}

    if((sleep_emu||pause_emu)&&needsredraw) {XCopyArea(display,layout,w,gc,0,0,256*magstep,240*magstep,0,0);needsredraw=0;}
  }

  if (needsredraw) {
#ifdef X11_DRAW
    XCopyArea(display,layout,w,gc,0,0,256*magstep,240*magstep,0,0);
    XFlush(display);
    /* 
       The purpose of this is to wait for the acknowledgement from the
       X server (a NoExpose event on the window background in response
       to the CopyArea) before proceeding, so we don't send commands
       faster than the X server can process them.  The -1 acts as a flag
       to indicate that the XEvent loop should block.
    */
    ev.type=-1;
#endif
  }

#ifdef X11_DRAW
  
  /* See if the window size changed */
  XGetGeometry(display,w,&root,&junk,&junk,&width,&height,&junk,&junk);
  if(width!=oldwidth||height!=oldheight)
  {
    if((magstep=((width*240 > height*256) ? width/256 : height/240))<1) magstep=1;
    else if(magstep>maxsize) magstep=maxsize;
    oldwidth=width;oldheight=height;
    memset(displaycachevalid[currentcache],0,3840);
    memset(bitplanecachevalid[currentcache],0,3840);
    for(x=0;x<256;x++) tilechanged[x]=tiledirty[currentcache][x]=1;
    desync=1;
    XFillRectangle(display,background[currentcache],blackgc,0,0,256*magstep,240*magstep);
    XFillRectangle(display,bgplane1[currentcache],blackplanegc,0,0,256*magstep,240*magstep);
    XFillRectangle(display,bgplane2[currentcache],blackplanegc,0,0,256*magstep,240*magstep);
    XFillRectangle(display,bgplane3[currentcache],blackplanegc,0,0,256*magstep,240*magstep);
    XFillRectangle(display,bgmask[currentcache],blackplanegc,0,0,256*magstep,240*magstep);
  }

  XFlush(display);
#endif
  needsredraw=0;
  redrawbackground=0;
  redrawall=0;
  
  /* Check the time.  If we're getting behind, skip next frame to stay in sync. */
#ifdef SPEED_UP_IF_BEHIND
  gettimeofday(&time,NULL);
  timeframe=(time.tv_sec-basetime)*60+time.tv_usec/16666; /* NTSC */
  if(halfspeed) timeframe>>=1;
  if(doublespeed==2) timeframe<<=1;
  else if(doublespeed>2) timeframe*=doublespeed;
  if(frame>=timeframe||frame%20==0) frameskip=0;
  else frameskip=1;  
#endif
  
}


/* FIXME Palette array is updated _every frame_ even if palette didn't change */
/* Plus more stuff should be ripped out and cleaned up */
/* FIXME Palette array isn't updated until end of frame (after endclock = 81840) from
   update_display */
void update_colors()
{
  int x;
  
  if(indexedcolor)
  {
  }
  else /* truecolor */
  {
    if(depth==16) ((short *)palette)[24]=palette_16[VRAM[0x3f00]&63];
    else if(depth==24) ((int *)palette)[24]=palette_24[VRAM[0x3f00]&63];
  }

  /* Set pallette tables */
  if(indexedcolor)
  {
    /* Already done in InitDisplay */
  }
  else /* truecolor */
  {
    if(depth==16)
    {
      for(x=0;x<24;x++)
      {
        ((short *)palette)[x]=palette_16[VRAM[0x3f01+x+(x/3)]&63];
		/* This adds an additional +1 offset after each 3 entries--
		 * because 3f00 is mirrored to 3f04, 3f08, ... 3f1c. */
      }
    }
    else if(depth==24)
    {
      for(x=0;x<24;x++)
      {
        ((int *)palette)[x]=palette_24[VRAM[0x3f01+x+(x/3)]&63];
      }
    }
  }
}

